/*-
 * Copyright (c) 2008  Peter Pentchev
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Ringlet: confget.c 2735 2008-10-14 11:28:19Z roam $
 */

#define _GNU_SOURCE

#include <ctype.h>
#include <err.h>
#include <fnmatch.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "confget.h"
#include "confget_ini.h"

/* The configuration filename specified by the user. */
const char		*filename;
/* The configuration section specified by the user. */
const char		*section;
/* The number of the variable or pattern parameters after the options. */
static int		 margc;
/* The variables to extract or patterns to match against. */
static char		* const *margv;
/* Display the variable names, or only the values? */
static int		 showvarname;
/* Look for more variables after a match is found? */
static int		 manyvars;
/* The pattern to match the variable value */
static const char	*matchvalue;

/* Scan for more variables after one was found? */
int			 morevars;

static confget_backend	*backends[] = {
	&confget_ini_backend,
};
#define BACKEND_CNT	(sizeof(backends) / sizeof(backends[0]))

static int		 find_backend(const char *name);
static int		 init(int _argc, char * const *_argv);
static void		 usage(int _error);
static void		 version(void);

/* The currently used configuration backend (ini, jprop, etc.) */
confget_backend		*backend;

/* The currently processed input file */
FILE			*fp;

/* Has the "-c" (check only) option been specified? */
int			 cflag;
/* In check-only mode, has the variable been found? */
int			 cfound;
/* Has the "-l" (list the section) option been specified? */
static int		 lflag;
/* Has the "-L" (match varname against patterns) option been specified? */
static int		 Lflag;

/***
 * Function:
 *	usage		- display program usage information and exit
 * Inputs:
 *	error		- is this invoked in error?
 * Returns:
 *	Nothing.
 * Modifies:
 *	Writes the usage information to the standard output or standard
 *	error depending on the "error" flag.
 */
void
usage(int error)
{
	const char *s = "Usage:\n"
	    "confget [-c] [-N | -n] [-m pattern] [-s section] [-t type]"
	    " -f filename var...\n"
	    "confget [-N | -n] [-m pattern] [-s section] [-t type]"
	    " -f filename -L pattern...\n"
	    "confget [-N | -n] [-m pattern] [-s section] [-t type]"
	    " -f filename -l\n"
	    "confget [-hTV]\n"
	    "\t-c\tcheck if the variable is defined in the file;\n"
	    "\t-f\tspecify the configuration file to read from;\n"
	    "\t-h\tdisplay usage information and exit;\n"
	    "\t-L\tspecify which variables to display;\n"
	    "\t-l\tlist all variables in the specified section;\n"
	    "\t-m\tonly display values that match the specified pattern;\n"
	    "\t-N\talways display the variable name;\n"
	    "\t-n\tnever display the variable name;\n"
	    "\t-s\tspecify the configuration section to read;\n"
	    "\t-T\tlist the available configuration file types;\n"
	    "\t-t\tspecify the configuration file type;\n"
	    "\t-V\tdisplay program version information and exit.";

	if (error)
		errx(1, "%s", s);
	printf("%s\n", s);
}

/***
 * Function:
 *	version		- display program version information
 * Inputs:
 *	None.
 * Returns:
 *	Nothing.
 * Modifies:
 *	Nothing; displays program version information on the standard output.
 */
void
version(void)
{
	printf("confget 0.02\n");
}

/***
 * Function:
 *	init		- parse the command-line options
 * Inputs:
 *	argc, argv	- the command-line parameter
 * Returns:
 *	0 on success, -1 on error.
 * Modifies:
 *	Sets section and varname on success.
 *	Writes to the standard error on error.
 */
static int
init(int argc, char * const *argv)
{
	int bidx, ch, do_help, do_list, do_version, show_name, hide_name;
	size_t i;

	fp = NULL;

	cflag = Lflag = lflag = showvarname = manyvars = 0;
	do_help = do_list = do_version = show_name = hide_name = 0;
	matchvalue = NULL;
	bidx = find_backend("ini");
	if (bidx == -1) {
		fputs("INTERNAL ERROR: no 'ini' backend\n", stderr);
		return (-1);
	}
	while ((ch = getopt(argc, argv, "cf:hLlm:Nns:Tt:V")) != EOF) {
		switch (ch) {
			case 'c':
				cflag = 1;
				manyvars = 0;
				break;

			case 'f':
				filename = optarg;
				break;

			case 'h':
				do_help = 1;
				break;

			case 'L':
				Lflag = 1;
				showvarname = 1;
				manyvars = 1;
				break;

			case 'l':
				lflag = 1;
				showvarname = 1;
				manyvars = 1;
				break;

			case 'm':
				matchvalue = optarg;
				break;

			case 'N':
				show_name = 1;
				break;

			case 'n':
				hide_name = 1;
				break;

			case 's':
				section = optarg;
				break;

			case 'T':
				do_list = 1;
				break;

			case 't':
				bidx = find_backend(optarg);
				if (bidx == -1) {
					fprintf(stderr, "No backend '%s'\n",
					    optarg);
					return (-1);
				}
				break;
			case 'V':
				do_version = 1;
				break;

			case '?':
			default:
				usage(1);
				break;
		}
	}

	if (do_help)
		usage(0);
	if (do_version)
		version();
	if (do_list) {
		printf("Supported backends:");
		for (i = 0; i < BACKEND_CNT; i++)
			printf(" %s", backends[i]->name);
		puts("");
	}
	if (do_help || do_list || do_version)
		exit(0);

	argc -= optind;
	argv += optind;
	if (filename == NULL) {
		warnx("No configuration file name specified");
		return (-1);
	}
	if (cflag && (lflag || Lflag)) {
		warnx("The -c flag may not be used with -l or -L");
		return (-1);
	}
	if (argc == 0 && !lflag) {
		warnx("No matching criteria specified");
		return (-1);
	} else if ((argc > 0) && lflag) {
		warnx("Too many matching criteria specified");
		return (-1);
	}
	if (argc > 1 && !lflag) {
		showvarname = 1;
		manyvars = 1;
	}
	if (show_name) {
		if (hide_name) {
			warnx("The -N and -n flags may not be used together");
			return (-1);
		}
		showvarname = 1;
	} else if (hide_name) {
		showvarname = 0;
	}
	margc = argc;
	margv = argv;
	backend = backends[bidx];
	return (0);
}

/***
 * Function:
 *	find_backend		- find a confget backend by name
 * Inputs:
 *	name			- the name of the backend
 * Returns:
 *	The index in backends[] if found, -1 otherwise.
 * Modifies:
 *	Nothing.
 */
static int
find_backend(const char *name)
{
	size_t i, len;
	int tentative;
	
	len = strlen(name);
	for (i = 0, tentative = -1; i < BACKEND_CNT; i++) {
		if (strncmp(name, backends[i]->name, len))
			continue;
		if (backends[i]->name[len] == '\0')
			return (i);
		if (tentative >= 0)
			return (-1);
		tentative = i;
	}
	return (tentative);
}

/* FIXME: The following three functions REALLY belong in confget_ini.c */

/***
 * Function:
 *	openfile		- open the requested file for reading
 * Inputs:
 *	None.
 * Returns:
 *	0 on success, -1 on error.
 * Modifies:
 *	Stores the opened file into fp.
 */
static int
openfile(void)
{
	if (backend->openfile == NULL) {
		fprintf(stderr,
		    "INTERNAL ERROR: backend '%s' does not define a openfile routine\n",
		    backend->name);
		return (-1);
	}
	return (backend->openfile());
}

/***
 * Function:
 *	readfile		- scan an INI file for the requested variable
 * Inputs:
 *	None.
 * Returns:
 *	0 on success, -1 on error.
 * Modifies:
 *	May write to standard output if the variable has been found.
 */
static int
readfile(void)
{

	if (backend->readfile == NULL) {
		fprintf(stderr,
		    "INTERNAL ERROR: backend '%s' does not define a readfile routine\n",
		    backend->name);
		return (-1);
	}
	return (backend->readfile());
}

/***
 * Function:
 *	closefile		- close a scanned INI file
 * Inputs:
 *	None.
 * Returns:
 *	0 on success, -1 on error.
 * Modifies:
 *	Closes the file pointed to by fp.
 */
static int
closefile(void)
{

	if (backend->closefile == NULL) {
		fprintf(stderr,
		    "INTERNAL ERROR: backend '%s' does not define a closefile routine\n",
		    backend->name);
		return (-1);
	}
	return (backend->closefile());
}

/***
 * Function:
 *	foundvar	- process the user-requested variable
 * Inputs:
 *	sec		- the section the variable was found in
 *	name		- the variable name
 *	value		- the variable value
 * Returns:
 *	0 on success, -1 on failure.
 * Modifies:
 *	In check-only mode, sets cfound.
 *	In normal mode, writes to standard output.
 */
int
foundvar(const char *sec __unused, const char *name,
	const char *value)
{
	int i;

	if (!lflag) {
		for (i = 0; i < margc; i++)
			if (Lflag) {
				if (fnmatch(margv[i], name, 0) == 0)
					break;
			} else {
				if (!strcmp(name, margv[i]))
					break;
			}
		if (i == margc) {
			morevars = 1;
			return (0);
		}
	}

	if (matchvalue != NULL) {
		if (fnmatch(matchvalue, value, 0) == FNM_NOMATCH) {
			morevars = 1;
			return (0);
		}
	}
	if (cflag) {
		cfound = 1;
		morevars = 0;
	} else {
		if (showvarname)
			printf("%s=%s\n", name, value);
		else
			printf("%s\n", value);
		morevars = manyvars;
	}
	return (0);
}

/***
 * Main routine
 */
int
main(int argc, char * const *argv)
{

	if (init(argc, argv) == -1)
		return (1);

	cfound = 0;
	if (openfile() == -1)
		return (1);
	if (readfile() == -1) {
		closefile();
		return (1);
	}
	if (closefile() == -1)
		return (1);

	if (cflag)
		return (!cfound);
	return (0);
}
